// *** THIS FILE IS GENERATED - DO NOT EDIT ***
// See valid_enum_values_generator.py for modifications

/***************************************************************************
 *
 * Copyright (c) 2015-2025 The Khronos Group Inc.
 * Copyright (c) 2015-2025 Valve Corporation
 * Copyright (c) 2015-2025 LunarG, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ****************************************************************************/

// NOLINTBEGIN
template <>
ValidValue stateless::Context::IsValidEnumValue(VkPipelineCacheHeaderVersion value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkImageLayout value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkObjectType value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkFormat value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkImageTiling value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkImageType value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkQueryType value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkSharingMode value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkComponentSwizzle value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkImageViewType value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkBlendFactor value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkBlendOp value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkCompareOp value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkDynamicState value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkFrontFace value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkVertexInputRate value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkPrimitiveTopology value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkPolygonMode value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkStencilOp value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkLogicOp value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkBorderColor value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkFilter value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkSamplerAddressMode value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkSamplerMipmapMode value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkDescriptorType value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkAttachmentLoadOp value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkAttachmentStoreOp value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkPipelineBindPoint value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkCommandBufferLevel value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkIndexType value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkSubpassContents value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkTessellationDomainOrigin value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkSamplerYcbcrModelConversion value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkSamplerYcbcrRange value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkChromaLocation value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkDescriptorUpdateTemplateType value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkSamplerReductionMode value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkSemaphoreType value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkPipelineRobustnessBufferBehavior value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkPipelineRobustnessImageBehavior value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkQueueGlobalPriority value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkLineRasterizationMode value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkPresentModeKHR value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkColorSpaceKHR value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkFragmentShadingRateCombinerOpKHR value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkVideoEncodeTuningModeKHR value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkComponentTypeKHR value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkVideoEncodeAV1PredictionModeKHR value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkVideoEncodeAV1RateControlGroupKHR value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkTimeDomainKHR value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkDebugReportObjectTypeEXT value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkRasterizationOrderAMD value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkShaderInfoTypeAMD value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkValidationCheckEXT value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkDisplayPowerStateEXT value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkDeviceEventTypeEXT value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkDisplayEventTypeEXT value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkViewportCoordinateSwizzleNV value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkDiscardRectangleModeEXT value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkConservativeRasterizationModeEXT value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkBlendOverlapEXT value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkCoverageModulationModeNV value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkShadingRatePaletteEntryNV value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkCoarseSampleOrderTypeNV value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkRayTracingShaderGroupTypeKHR value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkGeometryTypeKHR value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkAccelerationStructureTypeKHR value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkCopyAccelerationStructureModeKHR value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkAccelerationStructureMemoryRequirementsTypeNV value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkMemoryOverallocationBehaviorAMD value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkPerformanceConfigurationTypeINTEL value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkQueryPoolSamplingModeINTEL value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkPerformanceOverrideTypeINTEL value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkPerformanceParameterTypeINTEL value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkValidationFeatureEnableEXT value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkValidationFeatureDisableEXT value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkCoverageReductionModeNV value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkProvokingVertexModeEXT value) const;
#ifdef VK_USE_PLATFORM_WIN32_KHR
template <>
ValidValue stateless::Context::IsValidEnumValue(VkFullScreenExclusiveEXT value) const;
#endif  // VK_USE_PLATFORM_WIN32_KHR
template <>
ValidValue stateless::Context::IsValidEnumValue(VkIndirectCommandsTokenTypeNV value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkDepthBiasRepresentationEXT value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkFragmentShadingRateTypeNV value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkFragmentShadingRateNV value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkAccelerationStructureMotionInstanceTypeNV value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkDeviceFaultAddressTypeEXT value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkDeviceFaultVendorBinaryHeaderVersionEXT value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkDeviceAddressBindingTypeEXT value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkMicromapTypeEXT value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkBuildMicromapModeEXT value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkCopyMicromapModeEXT value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkAccelerationStructureCompatibilityKHR value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkAccelerationStructureBuildTypeKHR value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkRayTracingLssIndexingModeNV value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkRayTracingLssPrimitiveEndCapsModeNV value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkDirectDriverLoadingModeLUNARG value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkOpticalFlowPerformanceLevelNV value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkOpticalFlowSessionBindingPointNV value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkAntiLagModeAMD value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkAntiLagStageAMD value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkShaderCodeTypeEXT value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkDepthClampModeEXT value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkCooperativeVectorMatrixLayoutNV value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkLayerSettingTypeEXT value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkLatencyMarkerNV value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkOutOfBandQueueTypeNV value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkBlockMatchWindowCompareModeQCOM value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkCubicFilterWeightsQCOM value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkDisplaySurfaceStereoTypeNV value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkClusterAccelerationStructureTypeNV value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkClusterAccelerationStructureOpTypeNV value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkClusterAccelerationStructureOpModeNV value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkPartitionedAccelerationStructureOpTypeNV value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkIndirectExecutionSetInfoTypeEXT value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkIndirectCommandsTokenTypeEXT value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkBuildAccelerationStructureModeKHR value) const;
template <>
ValidValue stateless::Context::IsValidEnumValue(VkShaderGroupShaderKHR value) const;

// NOLINTEND
